"""
Edited from https://github.com/pytorch/examples/blob/master/vae/main.py
Reproduce experiments from Kool 2020 and Yin 2019
"""

from __future__ import print_function

from typing import List

import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F
from storch import deterministic, backward
import storch
from torch.distributions import Distribution


class VAE(nn.Module):
    def __init__(self, args):
        super(VAE, self).__init__()

        self.latents = args.latents
        self.samples = args.samples
        self.sampling_method = self.initialize_method(args)
        self.fc1 = nn.Linear(784, 512)
        self.fc2 = nn.Linear(512, 256)

        self.fc3, self.fc4 = self.initialize_param_layers(self.latents, 256)
        self.fc5 = nn.Linear(256, 512)
        self.fc6 = nn.Linear(512, 784)

        self.activation = lambda x: F.leaky_relu(x, negative_slope=0.1)

    def initialize_param_layers(
        self, latents: int, prev_layer: int
    ) -> (nn.Module, nn.Module):
        pass

    def initialize_method(self, args) -> storch.method.Method:
        pass

    @deterministic
    def encode(self, x):
        h1 = self.activation(self.fc1(x))
        h2 = self.activation(self.fc2(h1))
        return self.fc3(h2)

    @deterministic
    def decode(self, z):
        z = self.shape_latent(z, self.latents)
        h3 = self.activation(self.fc4(z))
        h4 = self.activation(self.fc5(h3))
        return self.fc6(h4).sigmoid()

    def forward(
        self, x: storch.Tensor
    ) -> (storch.Tensor, storch.Tensor, storch.Tensor):
        logits = self.encode(x)
        params = self.logits_to_params(logits, self.latents)
        var_posterior = self.variational_posterior(params)
        prior = self.prior(var_posterior)

        KLD = self.KLD(var_posterior, prior)
        storch.add_cost(KLD, "KL-divergence")
        z = self.sampling_method(var_posterior)
        return self.decode(z), KLD, z

    def prior(self, posterior: Distribution) -> Distribution:
        pass

    def variational_posterior(self, params) -> Distribution:
        pass

    def logits_to_params(self, logits: storch.Tensor, latents: int) -> storch.Tensor:
        return logits

    def shape_latent(self, z: storch.Tensor, latents: int) -> storch.Tensor:
        return z

    def KLD(self, var_posterior: Distribution, prior: Distribution) -> storch.Tensor:
        return torch.distributions.kl_divergence(var_posterior, prior).sum(-1)

    def name(self):
        return "vae"



from typing import Optional

import torch
import storch

from storch.method.sampling import SampleWithoutReplacementMethod, AncestralPlate


class ScoreFunctionWOR(SampleWithoutReplacementMethod):
    """
    Implement Buy 4 REINFORCE Samples, Get a Baseline for Free! https://openreview.net/pdf?id=r1lgTGL5DE
    Use biased=True for the biased normalized version which has lower variance.
    """

    EPS = 1e-8

    def __init__(
        self, plate_name: str, k: int, biased: bool = True, use_baseline: bool = True
    ):
        # Use k + 1 to be able to compute kappa, the k+1th perturbed log-prob
        super().__init__(plate_name, k + 1)
        self.biased = biased
        self.use_baseline = use_baseline

    def plate_weighting(
        self, tensor: storch.StochasticTensor, plate: storch.Plate
    ) -> Optional[storch.Tensor]:
        return self._compute_iw(plate, self.biased).detach()

    def _compute_iw(self, plate: AncestralPlate, biased: bool):
        # Compute importance weights. The kth sample has 0 weight, and is only used to compute the importance weights
        q = (
            1
            - (
                -(
                    plate.log_probs
                    - plate.perturb_log_probs._tensor[..., self.k - 1].unsqueeze(-1)
                ).exp()
            ).exp()
        ).detach()
        iw = plate.log_probs.exp() / (q + self.EPS)
        # It's probably not going to like this? Would iw._tensor work?
        # Set the weight of the kth sample (kappa) to 0.
        iw[..., self.k - 1] = 0.0
        if biased:
            WS = storch.sum(iw, plate).detach()
            return iw / WS
        return iw

    def adds_loss(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> bool:
        # We only want to add a loss on the stochastic tensor with the same plate as the cost node.
        # This is because the estimator computes the gradient with the respect to the JOINT log probability.
        # If we would have added the gradient for all stochastic tensors, these would just be duplicates of the same
        # loss being added (ie that gradient would be oversampled)
        for distr_plate in tensor.plates:
            if distr_plate.name == self.plate_name:
                for cost_plate in cost_node.plates:
                    if cost_plate.name == self.plate_name:
                        if cost_plate is distr_plate:
                            return True
                        return False
                raise ValueError(
                    "The given tensor contains an ancestral plate that the cost node doesn't have."
                )
        return False

    def estimator(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> Optional[storch.Tensor]:
        cost_plate = None
        for _p in cost_node.plates:
            if _p.name == self.plate_name:
                cost_plate = _p
                break
        if self.use_baseline:
            iw = self._compute_iw(cost_plate, biased=False)
            BS = storch.sum(iw * cost_node, cost_plate)
            probs = cost_plate.log_probs.exp()
            if self.biased:
                # Equation 11
                WS = storch.sum(iw, cost_plate)
                WiS = (WS - iw + probs).detach()
                diff_cost = cost_node - BS / WS
                return storch.sum(iw / WiS * diff_cost.detach(), cost_plate)
            else:
                # Equation 10
                weighted_cost = cost_node * (1 - probs + iw)
                diff_cost = weighted_cost - BS
                return storch.sum(iw * diff_cost.detach(), cost_plate)
        else:
            # Equation 9
            iw = self._compute_iw(cost_plate, self.biased)
            return storch.sum(cost_node.detach() * iw, self.plate_name)



import storch
import torch
from torch.distributions import Bernoulli, OneHotCategorical

expect = storch.Expect("x")
probs = torch.tensor([0.01, 0.01, 0.01, 0.01, 0.01, 0.95], requires_grad=True)
indices = torch.tensor([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
b = OneHotCategorical(probs=probs)
z = expect.sample(b)
c = (2.4 * z * indices).sum(-1)
storch.add_cost(c, "no_baseline_cost")

storch.backward()

expect_grad = z.grad["probs"].clone()

method = storch.ScoreFunction("x", 1, baseline_factory="none")
# method = storch.REBAR()
grads = []
for i in range(100):
    b = OneHotCategorical(probs=probs)
    z = method.sample(b)
    c = (2.4 * z * indices).sum(-1) + 100
    storch.add_cost(c, "baseline_cost")

    storch.backward()
    grad = z.grad["probs"].clone()
    grads.append(grad)
grad_samples = storch.gather_samples(grads, "variance")
mean = storch.reduce_plates(grad_samples, plate_names=["variance"])
print("mean grad", mean)
print("expected grad", expect_grad)
print("specific_diffs", (mean - expect_grad) ** 2)
mse = storch.reduce_plates((grad_samples - expect_grad) ** 2).sum()
print("MSE", mse)
bias = (storch.reduce_plates((mean - expect_grad) ** 2)).sum()
print("bias", bias)


# That works... Adding constants to the costs doesn't change the gradient in expectation.



"""
Based on Liu et al 2019, Kool et al 2020
"""

import torch
import storch
from torch.distributions import Bernoulli
from storch.method import ScoreFunction, Expect, UnorderedSetEstimator

p = torch.tensor([0.6, 0.51, 0.48])
eta = torch.tensor(-0.0, requires_grad=True)
optim = torch.optim.SGD([eta], lr=1.0)


def experiment(method):
    optim.zero_grad()
    b = Bernoulli(logits=eta.repeat(3))
    x = method(b)
    cost = torch.sum((x - p) ** 2, -1)
    storch.add_cost(cost, "cost")
    storch.backward()
    return eta.grad.clone()


true_grad = experiment(Expect("x"))
print("True gradient", true_grad)

for n in range(2, 9):
    # _method = ScoreFunction("x", n, baseline_factory="batch_average")
    _method = UnorderedSetEstimator("x", n)
    gradients = []
    for i in range(1000):
        gradients.append(experiment(_method).unsqueeze(0))
    gradients = torch.cat(gradients)
    print(n)
    print("variance", gradients.var(0))
    print("bias", torch.sum((gradients.mean(0) - true_grad) ** 2))



from __future__ import annotations
from typing import Optional, Union, List

import storch
import torch
from torch.distributions import Distribution, Gumbel
import itertools
from storch.method.method import Method


class SampleWithoutReplacementMethod(Method):
    def __init__(self, plate_name: str, k: int):
        super().__init__(plate_name)
        if k < 2:
            raise ValueError(
                "Can only sample with replacement for more than 1 samples."
            )
        self.k = k
        self.reset()

    def reset(self):
        # Cumulative log probabilities of the samples
        self.joint_log_probs = None
        # Cumulative perturbed log probabilities of the samples
        self.perturbed_log_probs = None
        # Chosen parent samples at the previous sample step
        self.parent_indexing = None
        # The index of the currently sampled variable
        self.variable_index = 0
        # The previously created plates
        self.last_plate = None

    def _sample_tensor(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [storch.Plate],
        requires_grad: bool,
    ) -> (torch.Tensor, storch.Plate):
        # Perform stochastic beam search given the log-probs and perturbed log probs so far.
        # Samples k values from this distribution so that all total configurations are unique.

        multi_dim_plates = []
        for plate in plates:
            if plate.n > 1:
                multi_dim_plates.append(plate)
        # Sample using stochastic beam search
        # plates? x k x events x
        samples = self.stochastic_beam_search(distr, multi_dim_plates)

        k_index = 0
        if isinstance(samples, storch.Tensor):
            k_index = samples.plate_dims
            plates = samples.plates
            samples = samples._tensor

        plate_size = samples.shape[k_index]

        # Remove the ancestral plate, if it already happens to be in
        to_remove = None
        for plate in plates:
            if plate.name == self.plate_name:
                to_remove = plate
                break
        if to_remove:
            plates.remove(to_remove)

        # Create the newly updated plate
        self.last_plate = AncestralPlate(
            self.plate_name,
            plate_size,
            plates.copy(),
            self.variable_index,
            self.last_plate,
            self.parent_indexing,
            self.joint_log_probs,
            self.perturbed_log_probs,
            None,
        )
        plates.append(self.last_plate)

        if self.parent_indexing is not None:
            self.parent_indexing.plates.append(self.last_plate)

        # Construct the stochastic tensor
        s_tensor = storch.StochasticTensor(
            samples,
            parents,
            plates,
            self.plate_name,
            plate_size,
            self,
            distr,
            requires_grad or self.joint_log_probs.requires_grad,
        )
        # Increase variable index
        self.variable_index += 1
        return s_tensor, self.last_plate

    def stochastic_beam_search(
        self, distribution: Distribution, orig_distr_plates: [storch.Plate],
    ) -> torch.Tensor:
        """
        Implements Ancestral Gumbel-Top-k sampling with m=k, known as Stochastic Beam Search: http://jmlr.org/papers/v21/19-985.html.
        Sample k events from the distribution so that the total generated sample throughout all sample steps
        is a sample without replacement. This method supports sampling without replacement on  arbitrary stochastic
        computation graphs.
        :param distribution: The distribution to sample from
        :return:
        """

        # This code has three parts.
        # The first prepares all necessary tensors to make sure they can be easily indexed.
        # This part is quite long as there are many cases.
        # 1) There have not been any variables sampled so far.
        # 2) There have been variables sampled, but their results are NOT used to compute the input distribution.
        #    in other words, the variable to sample is independent of the other sampled variables. However,
        #    we should still keep track of the other sampled variables to make sure that it still samples without
        #    replacement properly. In this case, the ancestral plate is not in the plates attribute.
        #    We also have to make sure that we know in the future what samples are chosen for the _other_ samples.
        # 3) There have been parents sampled, and this variable is dependent on at least some of them.
        #    The plates list then contains the ancestral plate. We need to make sure we compute the joint log probs
        #    for the conditional samples (ie, based on the different sampled variables in the ancestral dimension).
        # The second part is a loop over all options for the event dimensions. This samples these conditionally
        # independent samples in sequence. It samples indexes, not events.
        # The third part after the loop uses the sampled indexes and matches it to the events to be used.

        # LEGEND FOR SHAPE COMMENTS
        # =========================
        # To make this code generalize to every bayesian network, complicated shape management is necessary.
        # The following are references to the shapes that are used within the method
        #
        # distr_plates: refers to the plates on the parameters of the distribution. Does *not* include
        #  the k? ancestral plate (possibly empty)
        # orig_distr_plates: refers to the plates on the parameters of the distribution, and *can* include
        #  the k? ancestral plate (possibly empty)
        # prev_plates: refers to the plates of the previous sampled variable in this swr sample (possibly empty)
        # plates: refers to all plates except this ancestral plate, of which there are amt_plates.
        #  It is composed of distr_plate x (ancstr_plates - distr_plates)
        # events: refers to the conditionally independent dimensions of the distribution (the distributions batch shape minus the plates)
        # k: refers to self.k
        # k?: refers to an optional plate dimension of this ancestral plate. It either doesn't exist, or is the sample
        #  dimension. If it exists, this means this sample is conditionally dependent on ancestors.
        # |D_yv|: refers to the *size* of the domain
        # amt_samples: refers to the current amount of sampled sequences. amt_samples <= k, but it can be lower if there
        #  are not enough events to sample from (eg |D_yv| < k)
        # event_shape: refers to the *shape* of the domain elements
        #  (can be 0, eg Categorical, or equal to |D_yv| for OneHotCategorical)

        ancestral_distrplate_index = -1
        is_conditional_sample = False
        distr_plates = orig_distr_plates
        for i, plate in enumerate(orig_distr_plates):
            if plate.name == self.plate_name:
                ancestral_distrplate_index = i
                is_conditional_sample = True
                distr_plates = orig_distr_plates.copy()
                distr_plates.remove(plate)
                break

        # TODO: This doesn't properly combine two ancestral plates with the same name but different variable index
        #  (they should merge).
        all_plates = distr_plates.copy()
        prev_plate_shape = ()
        if isinstance(self.joint_log_probs, storch.Tensor):
            # Previous variables have been sampled. add the prev_plates to all_plates
            for plate in self.joint_log_probs.plates:
                if plate not in distr_plates:
                    all_plates.append(plate)
                    prev_plate_shape = prev_plate_shape + (plate.n,)

        amt_plates = len(all_plates)
        amt_distr_plates = len(distr_plates)
        amt_orig_distr_plates = len(orig_distr_plates)
        if not distribution.has_enumerate_support:
            raise ValueError(
                "Can only perform stochastic beam search for distributions with enumerable support."
            )

        # |D_yv| x distr_plate[0] x ... x k? x ... x distr_plate[n-1] x events x event_shape
        support = distribution.enumerate_support(expand=True)

        if ancestral_distrplate_index != -1:
            # Reduce ancestral dimension in the support. As the dimension is just an expanded version, this should
            # not change the underlying data.
            # |D_yv| x distr_plates x events x event_shape
            support = support[
                (..., 0)
                + (slice(None),) * (len(support.shape) - ancestral_distrplate_index - 2)
            ]

        # Equal to event_shape
        element_shape = distribution.event_shape
        support_permutation = (
            tuple(range(1, amt_distr_plates + 1))
            + (0,)
            + tuple(range(amt_distr_plates + 1, len(support.shape)))
        )
        # distr_plates x |D_yv| x events x event_shape
        support = support.permute(support_permutation)

        if amt_plates != amt_distr_plates:
            # If previous samples had a plate that are not in the distribution plates, add these to the support.
            support = support[
                (slice(None),) * amt_distr_plates
                + (None,) * (amt_plates - amt_distr_plates)
            ]
            # plates x |D_yv| x events x event_shape
            support = support.expand(
                (-1,) * amt_distr_plates
                + prev_plate_shape
                + (-1,) * (len(support.shape) - amt_distr_plates - 1)
            )
        support = storch.Tensor(support, [], all_plates)

        support_k_index = amt_plates

        # Equal to events: Shape for the different conditional independent dimensions
        events = support.shape[
            amt_plates + 1 : -len(element_shape) if len(element_shape) > 0 else None
        ]

        ranges = []
        for size in events:
            ranges.append(list(range(size)))

        with storch.ignore_wrapping():
            # |D_yv| x (|distr_plates| + |k?| + |event_dims|) * (1,) x |D_yv|
            support_non_expanded: torch.Tensor = distribution.enumerate_support(
                expand=False
            )
            # Compute the log-probability of the different events
            # |D_yv| x distr_plate[0] x ... k? ... x distr_plate[n-1] x events
            d_log_probs = distribution.log_prob(support_non_expanded)

            # Note: Use amt_orig_distr_plates here because it might include k? dimension. amt_distr_plates filters this one.
            # distr_plate[0] x ... k? ... x distr_plate[n-1] x |D_yv| x events
            d_log_probs = storch.Tensor(
                d_log_probs.permute(
                    tuple(range(1, amt_orig_distr_plates + 1))
                    + (0,)
                    + tuple(
                        range(
                            amt_orig_distr_plates + 1,
                            amt_orig_distr_plates + 1 + len(events),
                        )
                    )
                ),
                [],
                orig_distr_plates,
            )
        if is_conditional_sample:
            # Gather the correct log probabilities
            # distr_plate[0] x ... k ... x distr_plate[n-1] x |D_yv| x events
            # TODO: Move this down below to the other scary TODO
            d_log_probs = self.last_plate.on_unwrap_tensor(d_log_probs)
            # Permute the dimensions of d_log_probs st the k dimension is after the plates.
            for i, plate in enumerate(d_log_probs.multi_dim_plates()):
                if plate.name == self.plate_name:
                    # k is present in the plates
                    d_log_probs.plates.remove(plate)
                    # distr_plates x k x |D_yv| x events
                    d_log_probs._tensor = d_log_probs._tensor.permute(
                        tuple(range(0, i))
                        + tuple(range(i + 1, amt_orig_distr_plates))
                        + (i,)
                        + tuple(range(amt_orig_distr_plates, len(d_log_probs.shape)))
                    )
                    break

        # plates x k x events
        sampled_support_indices = support.new_zeros(
            size=support.shape[:support_k_index]  # distr_plates
            + (self.k,)
            + support.shape[
                support_k_index + 1 : -len(element_shape)
                if len(element_shape) > 0
                else None
            ],  # events
            dtype=torch.long,
        )

        amt_samples = 0
        self.parent_indexing = None
        if self.joint_log_probs is not None:
            # Initialize a tensor (self.parent_indexing) that keeps track of what samples link to previous choices of samples
            # Note that self.joint_log_probs.shape[-1] is amt_samples, not k. It's possible that amt_samples < k!
            amt_samples = self.joint_log_probs.shape[-1]
            # plates x k
            self.parent_indexing = support.new_zeros(
                size=support.shape[:amt_plates] + (self.k,), dtype=torch.long
            )

            # probably can go wrong if plates are missing.
            self.parent_indexing[..., :amt_samples] = left_expand_as(
                torch.arange(amt_samples), self.parent_indexing
            )

        # Sample independent tensors in sequence
        # Iterate over the different (conditionally) independent samples being taken (the events)
        for indices in itertools.product(*ranges):
            # Log probabilities of the different options for this sample step (event)
            # distr_plates x k? x |D_yv|
            yv_log_probs = d_log_probs[(...,) + indices]
            first_sample = False
            if self.joint_log_probs is None:
                # We also know that k? is not present, so distr_plates x |D_yv|
                all_joint_log_probs = yv_log_probs
                # First condition on max being 0:
                self.perturbed_log_probs = 0.0
                first_sample = True
            elif is_conditional_sample:
                # Make sure we are selecting the correct log-probabilities. As parents have been selected, this might change!
                # plates x amt_samples x |D_yv|
                yv_log_probs = yv_log_probs.gather(
                    dim=-2,
                    index=right_expand_as(
                        # Use the parent_indexing to select the correct plate samples. Make sure to limit to amt_samples!
                        self.parent_indexing[..., :amt_samples],
                        yv_log_probs,
                    ),
                )
                # self.joint_log_probs: prev_plates x amt_samples
                # plates x amt_samples x |D_yv|
                all_joint_log_probs = self.joint_log_probs.unsqueeze(-1) + yv_log_probs
            else:
                # self.joint_log_probs: plates x amt_samples
                # plates x amt_samples x |D_yv|
                all_joint_log_probs = self.joint_log_probs.unsqueeze(
                    -1
                ) + yv_log_probs.unsqueeze(-2)

            # Sample plates x k? x |D_yv| conditional Gumbel variables
            cond_G_yv = cond_gumbel_sample(
                all_joint_log_probs, self.perturbed_log_probs
            )
            if first_sample:
                # No parent has been sampled yet
                # shape(cond_G_yv) is plates x |D_yv|
                amt_samples = min(self.k, cond_G_yv.shape[-1])
                # Compute top k over the conditional perturbed log probs
                # plates x amt_samples
                self.perturbed_log_probs, arg_top = torch.topk(
                    cond_G_yv, amt_samples, dim=-1
                )
                # plates x amt_samples
                self.joint_log_probs = all_joint_log_probs.gather(dim=-1, index=arg_top)
                # Index for the selected samples. Uses slice(amt_samples) for the first index in case k > |D_yv|
                # (:) * amt_plates + (indices for events) + amt_samples
                indexing = (
                    (slice(None),) * amt_plates + (slice(0, amt_samples),) + indices
                )
                sampled_support_indices[indexing] = arg_top
            else:
                # plates x (k * |D_yv|) (k == prev_amt_samples, in this case)
                cond_G_yv = cond_G_yv.reshape(cond_G_yv.shape[:-2] + (-1,))
                # We can sample at most the amount of what we previous sampled, combined with every option in the current domain
                # That is: prev_amt_samples * |D_yv|. But we also want to limit by k.
                amt_samples = min(self.k, cond_G_yv.shape[-1])
                # Take the top k over conditional perturbed log probs
                # plates x amt_samples
                self.perturbed_log_probs, arg_top = torch.topk(
                    cond_G_yv, amt_samples, dim=-1
                )
                # Gather corresponding joint log probabilities. First reshape like previous to plates x (k * |D_yv|).
                self.joint_log_probs = all_joint_log_probs.reshape(
                    cond_G_yv.shape
                ).gather(dim=-1, index=arg_top)

                # |D_yv|
                size_domain = yv_log_probs.shape[-1]

                # Keep track of what parents were sampled for the arg top
                # plates x amt_samples
                chosen_parents = arg_top // size_domain
                sampled_support_indices = sampled_support_indices.gather(
                    dim=support_k_index,
                    index=right_expand_as(chosen_parents, sampled_support_indices),
                )
                if self.parent_indexing is not None:
                    self.parent_indexing = self.parent_indexing.gather(
                        dim=-1, index=chosen_parents
                    )
                # Index for the selected samples. Uses slice(amt_samples) for the first index in case k > |D_yv|
                # plates x amt_samples
                chosen_samples = arg_top.remainder(size_domain)
                indexing = (
                    (slice(None),) * amt_plates + (slice(0, amt_samples),) + indices
                )
                sampled_support_indices[indexing] = chosen_samples

        # Finally, index the support using the sampled indices to get the sample!
        if amt_samples < self.k:
            # plates x amt_samples x events
            sampled_support_indices = sampled_support_indices[
                (...,) + (slice(amt_samples),) + (slice(None),) * len(events)
            ]
        expanded_indices = right_expand_as(sampled_support_indices, support)
        return support.gather(dim=support_k_index, index=expanded_indices)

    def on_plate_already_present(self, plate: storch.Plate):
        if (
            not isinstance(plate, AncestralPlate)
            or plate.variable_index > self.variable_index
            or plate.n > self.k
        ):
            super().on_plate_already_present(plate)


class AncestralPlate(storch.Plate):
    def __init__(
        self,
        name: str,
        n: int,
        parents: List[storch.Plate],
        variable_index: int,
        parent_plate: AncestralPlate,
        selected_samples: storch.Tensor,
        log_probs: storch.Tensor,
        perturb_log_probs: storch.Tensor,
        weight: Optional[storch.Tensor] = None,
    ):
        super().__init__(name, n, parents, weight)
        assert (not parent_plate and variable_index == 0) or (
            parent_plate.n <= self.n and parent_plate.variable_index < variable_index
        )
        self.parent_plate = parent_plate
        self.selected_samples = selected_samples
        self.log_probs = storch.Tensor(
            log_probs._tensor, [log_probs], log_probs.plates + [self]
        )
        self.perturb_log_probs = storch.Tensor(
            perturb_log_probs._tensor,
            [perturb_log_probs],
            perturb_log_probs.plates + [self],
        )
        self.variable_index = variable_index
        self._in_recursion = False
        self._override_equality = False

    def __eq__(self, other):
        if self._override_equality:
            return other.name == self.name
        return (
            super().__eq__(other)
            and isinstance(other, AncestralPlate)
            and self.variable_index == other.variable_index
        )

    def __repr__(self):
        return (
            "(Ancestral, " + self.variable_index.__repr__() + super().__repr__() + ")"
        )

    def on_collecting_args(self, plates: [storch.Plate]) -> bool:
        """
        Filter the collected plates to only keep the AncestralPlates (with the same name) that has the highest variable index.
        :param plates:
        :return:
        """
        if self._in_recursion:
            self._override_equality = True
        for plate in plates:
            if plate.name == self.name:
                if not isinstance(plate, AncestralPlate):
                    raise ValueError(
                        "Received a plate with name "
                        + plate.name
                        + " that is not also an AncestralPlate."
                    )
                if plate.variable_index > self.variable_index:
                    # Only keep ancestral plates with the highest variable index
                    return False
        return True

    def on_unwrap_tensor(self, tensor: storch.Tensor) -> storch.Tensor:
        """
        Gets called whenever the given tensor is being unwrapped and unsqueezed for batch use.
        This method should not be called on tensors whose variable index is higher than this plates.
        :param tensor: The input tensor that is being unwrapped
        :return: The tensor that will be unwrapped and unsqueezed in the future. Can be a modification of the input tensor.
        """
        if self._in_recursion:
            # Required when calling storch.gather in this method. It will call on_unwrap_tensor again.
            return tensor
        for i, plate in enumerate(tensor.multi_dim_plates()):
            if plate.name != self.name:
                continue
            assert isinstance(plate, AncestralPlate)
            if plate.variable_index == self.variable_index:
                return tensor
            # This is true by the filtering at on_collecting_args
            assert plate.variable_index < self.variable_index

            parent_plates = []
            current_plate = self

            # Collect the list of plates from the tensors variable index to this plates variable index
            while current_plate.variable_index != plate.variable_index:
                parent_plates.append(current_plate)
                current_plate = current_plate.parent_plate
            assert current_plate == plate

            # Go over all parent plates and gather their respective choices.
            for parent_plate in reversed(parent_plates):
                self._in_recursion = True
                expanded_selected_samples = expand_with_ignore_as(
                    parent_plate.selected_samples, tensor, self.name
                )
                self._override_equality = False
                # Gather what samples of the tensor are chosen by this plate (parent_plate)
                tensor = storch.gather(
                    tensor, parent_plate.name, expanded_selected_samples
                )
                self._in_recursion = False
                self._override_equality = False
            break
        return tensor


def log1mexp(a: torch.Tensor) -> torch.Tensor:
    """See appendix A of http://jmlr.org/papers/v21/19-985.html.
    Numerically stable implementation of log(1-exp(a))"""
    c = -0.693
    a1 = -a.abs()
    return torch.where(a1 > c, torch.log(-a1.expm1()), torch.log1p(-a1.exp()))


@storch.deterministic
def cond_gumbel_sample(all_joint_log_probs, perturbed_log_probs) -> torch.Tensor:
    # Sample plates x k? x |D_yv| Gumbel variables
    gumbel_d = Gumbel(loc=all_joint_log_probs, scale=1.0)
    G_yv = gumbel_d.rsample()

    # Condition the Gumbel samples on the maximum of previous samples
    # plates x k
    Z = G_yv.max(dim=-1)[0]
    T = perturbed_log_probs
    vi = T - G_yv + log1mexp(G_yv - Z.unsqueeze(-1))
    # plates (x k) x |D_yv|
    return T - vi.relu() - torch.nn.Softplus()(-vi.abs())


@storch.deterministic(l_broadcast=False)
def right_expand_as(tensor, expand_as):
    diff = expand_as.ndim - tensor.ndim
    return tensor[(...,) + (None,) * diff].expand(
        (-1,) * tensor.ndim + expand_as.shape[tensor.ndim :]
    )


def left_expand_as(tensor, expand_as):
    diff = expand_as.ndim - tensor.ndim
    return tensor[(None,) * diff].expand(expand_as.shape[:diff] + (-1,) * tensor.ndim)


def expand_with_ignore_as(
    tensor, expand_as, ignore_dim: Union[str, int]
) -> torch.Tensor:
    """
    Expands the tensor like expand_as, but ignores a single dimension.
    Ie, if tensor is of size a x b,  expand_as of size d x a x c and dim=-1, then the return will be of size d x a x b.
    It also automatically expands all plate dimensions correctly.
    :param ignore_dim: Can be a string referring to the plate dimension
    """
    # diff = expand_as.ndim - tensor.ndim
    def _expand_with_ignore(tensor, expand_as, dim: int):
        new_dims = expand_as.ndim - tensor.ndim
        # after_dims = tensor.ndim - dim
        return tensor[(...,) + (None,) * new_dims].expand(
            expand_as.shape[:dim]
            + (-1,)
            + (expand_as.shape[dim + 1 :] if dim != -1 else ())
        )

    if isinstance(ignore_dim, str):
        return storch.deterministic(
            _expand_with_ignore, expand_plates=True, dim=ignore_dim
        )(tensor, expand_as)
    return storch.deterministic(_expand_with_ignore, expand_plates=True)(
        tensor, expand_as, ignore_dim
    )



import torch

a = torch.tensor([[[[1.0, 0.4, 0.6, 0.9], [5.0, 4.0, 3.0, 0.2]]]])
b = torch.tensor([[[[True, False, True, False], [True, False, True, False]]]])

import storch

print(torch.Tensor.__getitem__(a, b))
print(a[b])
a[b] = 10.0
print(a)
from torch.distributions import Normal, Categorical, Poisson
from storch.exceptions import IllegalStorchExposeError

method = storch.method.Infer("oeps", Poisson)

d = Normal(0, 1)
s = method.sample(d)
# print(s>1)
try:
    if s:
        assert False
except IllegalStorchExposeError:
    pass

try:
    if s == s:
        assert False
except IllegalStorchExposeError:
    pass

d = Poisson(4)
s = method.sample(d).int()

print(s)

i = 0
try:
    for _ in range(s):
        i += 1
    assert False
except IllegalStorchExposeError:
    pass

try:
    if torch.allclose(s, s):
        assert False
except IllegalStorchExposeError:
    pass

try:
    if s.allclose(s):
        assert False
except IllegalStorchExposeError:
    pass



from __future__ import annotations

from typing import Union, Any, Tuple, List, Optional, Dict

import storch
import torch
from collections.abc import Iterable, Mapping
from functools import wraps
from storch.exceptions import IllegalStorchExposeError
from contextlib import contextmanager

_context_stochastic = False
_context_deterministic = 0
_stochastic_parents = []
_context_name = None
_plate_links = []
_ignore_wrap = False

# TODO: This is_iterable thing is a bit annoying: We really only want to unwrap them if they contain storch
#  Tensors, and then only for some types. Should rethink, maybe. Is unwrapping even necessary if the base torch methods
#  are all overriden? Maybe, see torch.cat?
def is_iterable(a: Any):
    return (
        isinstance(a, Iterable)
        and not storch.is_tensor(a)
        and not isinstance(a, str)
        and not isinstance(a, torch.Storage)
    )


def _collect_parents_and_plates(
    a: Any, parents: [storch.Tensor], plates: [storch.Plate]
) -> int:
    if isinstance(a, storch.Tensor):
        parents.append(a)
        for plate in a.plates:
            if plate not in plates:
                plates.append(plate)
        return a.event_dims
    elif isinstance(a, Mapping):
        max_event_dim = 0
        for _a in a.values():
            max_event_dim = max(
                max_event_dim, _collect_parents_and_plates(_a, parents, plates)
            )
        return max_event_dim
    elif is_iterable(a):
        max_event_dim = 0
        for _a in a:
            max_event_dim = max(
                max_event_dim, _collect_parents_and_plates(_a, parents, plates)
            )
        return max_event_dim
    return 0


def _unsqueeze_and_unwrap(
    a: Any,
    multi_dim_plates: [storch.Plate],
    align_tensors: bool,
    l_broadcast: bool,
    expand_plates: bool,
    event_dims: int,
):
    if isinstance(a, storch.Tensor):
        if not align_tensors:
            return a._tensor

        for plate in multi_dim_plates:
            # Used in storch.method.sampling.AncestralPlate
            a = plate.on_unwrap_tensor(a)

        tensor = a._tensor
        # Automatically **RIGHT** broadcast. Ensure each tensor has an equal amount of event dims by inserting dimensions to the right
        # TODO: What do we think about this design?
        # TODO: The storch.tensor._getitem_level == 0 check prevents right-broadcasting for __getitem__ and __setitem__... Seems hacky
        if l_broadcast and a.event_dims < event_dims:
            tensor = tensor[(...,) + (None,) * (event_dims - a.event_dims)]
        # It can be possible that the ordering of the plates does not align with the ordering of the inputs.
        # This part corrects this.
        amt_recognized = 0
        links: [storch.Plate] = a.multi_dim_plates()

        for i, plate in enumerate(multi_dim_plates):
            if plate in links:
                if plate != links[amt_recognized]:
                    # The plate is also in the tensor, but not in the ordering expected. So switch that ordering
                    j = links.index(plate)
                    tensor = tensor.transpose(j, amt_recognized)
                    links[amt_recognized], links[j] = links[j], links[amt_recognized]
                amt_recognized += 1

        # Add singleton dimensions on missing plates
        plate_dims = []
        for i, plate in enumerate(multi_dim_plates):
            if plate not in a.plates:
                tensor = tensor.unsqueeze(i)
                plate_dims.append(plate.n)
            else:
                # Make sure to use a's plate size here. It's actually possible they are different in ancestral plates!
                plate_dims.append(tensor.shape[i])

        # Optionally expand the singleton dimensions to the plate size
        if expand_plates:
            tensor = tensor.expand(tuple(plate_dims) + tensor.shape[len(plate_dims) :])
        return tensor
    elif isinstance(a, Mapping):
        d = {}
        for k, _a in a.items():
            d[k] = _unsqueeze_and_unwrap(
                _a,
                multi_dim_plates,
                align_tensors,
                l_broadcast,
                expand_plates,
                event_dims,
            )
        return d
    elif is_iterable(a):
        l = []
        for _a in a:
            l.append(
                _unsqueeze_and_unwrap(
                    _a,
                    multi_dim_plates,
                    align_tensors,
                    l_broadcast,
                    expand_plates,
                    event_dims,
                )
            )
        if isinstance(a, tuple):
            return tuple(l)
        return l
    else:
        return a


def _prepare_args(
    fn_args,
    fn_kwargs,
    unwrap=True,
    align_tensors=True,
    l_broadcast=True,
    expand_plates=False,
    dim: Optional[str] = None,
    dims: Optional[Union[str, List[str]]] = None,
) -> (List, Dict, [storch.Tensor], [storch.Plate]):
    """
    Prepares the input arguments of the wrapped function:
    - Unwrap the input arguments from storch.Tensors to normal torch.Tensors so they can be used in any torch function
    - Align plate dimensions for automatic broadcasting
    - Add (singleton) plate dimensions for plates that are not present
    - Right-broadcast event dimensions for automatic broadcasting
    - Superclasses of Plate specific input handling
    :param fn_args: List of non-keyword arguments to the wrapped function
    :param fn_kwargs: Dictionary of keyword arguments to the wrapped function
    :param unwrap: Whether to unwrap the arguments to their torch.Tensor counterpart (default: True)
    :param align_tensors: Whether to automatically align the input arguments (default: True)
    :param l_broadcast: Whether to automatically left-broadcast (default: True)
    :param expand_plates: Instead of adding singleton dimensions on non-existent plates, this will
    add the plate size itself (default: False)
    :param dim: Replaces the dim input in fn_kwargs by the plate dimension corresponding to the given string (optional)
    :param dims: Replaces the dims input in fn_kwargs by the plate dimensions corresponding to the given strings (optional)
    :param self_wrapper: storch.Tensor that wraps a
    :return: Handled non-keyword arguments, handled keyword arguments, list of parents, list of plates
    """
    parents: [storch.Tensor] = []
    plates: [storch.Plate] = []
    max_event_dim = max(
        # Collect parent tensors and plates
        _collect_parents_and_plates(fn_args, parents, plates),
        _collect_parents_and_plates(fn_kwargs, parents, plates),
    )

    # Allow plates to filter themselves from being collected. This is used in storch.method.sampling.AncestralPlate
    plates = list(filter(lambda p: p.on_collecting_args(plates), plates))

    # Get the list of plates with size larger than 1 for the unsqueezing of tensors
    multi_dim_plates = []
    for plate in plates:
        if plate.n > 1:
            multi_dim_plates.append(plate)

    if dim:
        for i, plate in enumerate(multi_dim_plates):
            if dim == plate.name:
                i_dim = i
                break
        fn_kwargs["dim"] = i_dim
    if dims:
        dimz = []
        for dim in dims:
            for i, plate in enumerate(multi_dim_plates):
                if dim == plate.name:
                    dimz.append(i_dim)
                    break
            raise ValueError("Missing plate dimension" + dim)
        fn_kwargs["dims"] = dimz

    if unwrap:
        # Unsqueeze and align batched dimensions so that batching works easily.
        unsqueezed_args = []
        for t in fn_args:
            unsqueezed_args.append(
                _unsqueeze_and_unwrap(
                    t,
                    multi_dim_plates,
                    align_tensors,
                    l_broadcast,
                    expand_plates,
                    max_event_dim,
                )
            )
        unsqueezed_kwargs = {}
        for k, v in fn_kwargs.items():
            unsqueezed_kwargs[k] = _unsqueeze_and_unwrap(
                v,
                multi_dim_plates,
                align_tensors,
                l_broadcast,
                expand_plates,
                max_event_dim,
            )
        return unsqueezed_args, unsqueezed_kwargs, parents, plates
    return fn_args, fn_kwargs, parents, plates


def _process_deterministic(
    o: Any, parents: [storch.Tensor], plates: [storch.Plate], name: str
):
    if o is None:
        return
    if isinstance(o, storch.Tensor):
        if o.stochastic:
            raise RuntimeError(
                "Creation of stochastic storch Tensor within deterministic context"
            )
        # TODO: Does this require shape checking? Parent/Plate checking?
        return o
    if isinstance(o, torch.Tensor):  # Explicitly _not_ a storch.Tensor
        t = storch.Tensor(o, parents, plates, name=name)
        return t
    raise NotImplementedError(
        "Handling of other types of return values is currently not implemented: ", o
    )


def _handle_deterministic(
    fn,
    fn_args,
    fn_kwargs,
    reduce_plates: Optional[Union[str, List[str]]] = None,
    **wrapper_kwargs
):
    if storch.wrappers._context_stochastic:
        raise NotImplementedError(
            "It is currently not allowed to open a deterministic context in a stochastic context"
        )
        # TODO check if we can re-add this
        # if storch.wrappers._context_deterministic > 0:
        #     if is_cost:
        #         raise RuntimeError("Cannot call storch.cost from within a deterministic context.")

        # TODO: This is currently uncommented and it will in fact unwrap. This was required because it was, eg,
        # possible to open a deterministic context, passing distributions with storch.Tensors as parameters,
        # then doing computations on these parameters. This is because these storch.Tensors will not be unwrapped
        # in the deterministic context as the unwrapping only considers lists.
        # # We are already in a deterministic context, no need to wrap or unwrap as only the outer dependencies matter
        # return fn(*args, **kwargs)

    new_fn_args, new_fn_kwargs, parents, plates = _prepare_args(
        fn_args, fn_kwargs, **wrapper_kwargs
    )
    if not parents:
        return fn(*fn_args, **fn_kwargs)
    args = new_fn_args
    kwargs = new_fn_kwargs

    storch.wrappers._context_deterministic += 1

    try:
        outputs = fn(*args, **kwargs)
    finally:
        storch.wrappers._context_deterministic -= 1

    if storch.wrappers._ignore_wrap:
        return outputs
    if reduce_plates:
        if isinstance(reduce_plates, str):
            reduce_plates = [reduce_plates]
        plates = [p for p in plates if p.name not in reduce_plates]
    if is_iterable(outputs):
        n_outputs = []
        for o in outputs:
            n_outputs.append(
                _process_deterministic(
                    o, parents, plates, fn.__name__ + str(len(n_outputs))
                )
            )
        outputs = n_outputs
    else:
        outputs = _process_deterministic(outputs, parents, plates, fn.__name__)
    return outputs


def _deterministic(
    fn, reduce_plates: Optional[Union[str, List[str]]] = None, **wrapper_kwargs
):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        nonlocal reduce_plates
        return _handle_deterministic(fn, args, kwargs, reduce_plates, **wrapper_kwargs)

    return wrapper


def deterministic(fn=None, **kwargs):
    """
    Wraps the input function around a deterministic storch wrapper.
    This wrapper unwraps :class:`~storch.Tensor` objects to :class:`~torch.Tensor` objects, aligning the tensors
    according to the plates, then runs `fn` on the unwrapped Tensors.
    :param fn: Function to wrap.
    :param unwrap: Set to False to prevent unwrapping :classs:`~storch.Tensor` objects.
    :return: The wrapped function `fn`.
    """
    if fn:
        return _deterministic(fn, **kwargs)
    return lambda _f: _deterministic(_f, **kwargs)


def reduce(fn, plates: Union[str, List[str]]):
    """
        Wraps the input function around a deterministic storch wrapper.
        This wrapper unwraps :class:`~storch.Tensor` objects to :class:`~torch.Tensor` objects, aligning the tensors
        according to the plates, then runs `fn` on the unwrapped Tensors. It will reduce the plates given by `plates`.
        :param fn: Function to wrap.
        :return: The wrapped function `fn`.
        """
    if storch._debug:
        print("Reducing plates", plates)
    return _deterministic(fn, reduce_plates=plates)


def _self_deterministic(fn, self: storch.Tensor):
    fn = deterministic(fn)

    @wraps(fn)
    def wrapper(*args, **kwargs):
        # Inserts the self object at the beginning of the passed arguments. In essence, it "fakes" the self reference.
        args = list(args)
        args.insert(0, self)
        return fn(*args, **kwargs)

    return wrapper


def _process_stochastic(
    output: torch.Tensor, parents: [storch.Tensor], plates: [storch.Plate]
):
    if isinstance(output, storch.Tensor):
        if not output.stochastic:
            # TODO: Calls _add_parents so something is going wrong here
            # The Tensor was created by calling @deterministic within a stochastic context.
            # This means that we have to conservatively assume it is dependent on the parents
            output._add_parents(storch.wrappers._stochastic_parents)
        return output
    if isinstance(output, torch.Tensor):
        t = storch.Tensor(output, parents, plates)
        return t
    else:
        raise TypeError(
            "All outputs of functions wrapped in @storch.stochastic "
            "should be Tensors. At " + str(output)
        )


def stochastic(fn):
    """
    Applies `fn` to the `inputs`. `fn` should return one or multiple `storch.Tensor`s.
    `fn` should not call `storch.stochastic` or `storch.deterministic`. `inputs` can include `storch.Tensor`s.
    :param fn:
    :return:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        if (
            storch.wrappers._context_stochastic
            or storch.wrappers._context_deterministic > 0
        ):
            raise RuntimeError(
                "Cannot call storch.stochastic from within a stochastic or deterministic context."
            )

        # Save the parents
        args, kwargs, parents, plates = _prepare_args(*args, **kwargs)

        storch.wrappers._plate_links = plates
        storch.wrappers._stochastic_parents = parents
        storch.wrappers._context_stochastic = True
        storch.wrappers._context_name = fn.__name__

        try:
            outputs = fn(*args, **kwargs)
        finally:
            storch.wrappers._plate_links = []
            storch.wrappers._stochastic_parents = []
            storch.wrappers._context_stochastic = False
            storch.wrappers._context_name = None

        # Add parents to the outputs
        if is_iterable(outputs):
            processed_outputs = []
            for o in outputs:
                processed_outputs.append(_process_stochastic(o, parents, plates))
        else:
            processed_outputs = _process_stochastic(outputs, parents, plates)
        return processed_outputs

    return wrapper


def _exception_wrapper(fn):
    def wrapper(*args, **kwargs):
        for a in args:
            if isinstance(a, storch.Tensor):
                raise IllegalStorchExposeError(
                    "It is not allowed to call this method using storch.Tensor, likely "
                    "because it exposes its wrapped tensor to Python."
                )
        return fn(*args, **kwargs)

    return wrapper


def _unpack_wrapper(fn):
    def wrapper(*args, **kwargs):
        new_args = []
        for a in args:
            if isinstance(a, storch.Tensor):
                new_args.append(a._tensor)
            else:
                new_args.append(a)
        return fn(*tuple(new_args), **kwargs)

    return wrapper


@contextmanager
def ignore_wrapping():
    storch.wrappers._ignore_wrap = True
    yield
    storch.wrappers._ignore_wrap = False



from functools import reduce
from operator import mul
from typing import Optional, Callable

import torch
from torch.distributions import Distribution, Bernoulli
from torch.distributions.utils import clamp_probs
from torch.nn import Parameter

import storch
from storch import Plate, CostTensor, StochasticTensor, deterministic
from storch.method.method import MonteCarloMethod
from storch.typing import Dims

import torch.nn.functional as F

from storch.util import get_distr_parameters, rsample_gumbel, split


class Baseline(torch.nn.Module):
    def __init__(self, in_dim: Dims):
        super().__init__()
        self.reshape = False
        if not isinstance(in_dim, int):
            self.reshape = True
            in_dim = reduce(mul, in_dim)

        self.fc1 = torch.nn.Linear(in_dim, 50)
        self.fc2 = torch.nn.Linear(50, 1)

    def forward(self, x: storch.Tensor):
        if self.reshape:
            x = x.reshape(x.shape[: x.plate_dims] + (-1,))
        return self.fc2(F.relu(self.fc1(x))).squeeze(-1)


class LAX(MonteCarloMethod):
    """
    Gradient estimator for continuous random variables.
    Implements the LAX estimator from Grathwohl et al, 2018 https://arxiv.org/abs/1711.00123
    Code inspired by https://github.com/duvenaud/relax/blob/master/pytorch_toy.py
    """

    def __init__(
        self,
        plate_name: str,
        *,
        n_samples: int = 1,
        c_phi: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        in_dim=None,
    ):
        super().__init__(plate_name, n_samples)
        if c_phi:
            self.c_phi = c_phi
        else:
            self.c_phi = Baseline(in_dim)
        # TODO: Add baseline strength

    def mc_sample(
        self, distr: Distribution, parents: [storch.Tensor], plates: [Plate]
    ) -> torch.Tensor:
        sample = distr.rsample((self.n_samples,))
        return sample

    def post_sample(self, tensor: storch.StochasticTensor) -> Optional[storch.Tensor]:
        return tensor.detach()

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Optional[storch.Tensor]:
        # Input rsampled value into c_phi
        output_baseline = self.c_phi(tensor)

        # Compute log probability. Make sure not to use the rsampled value: We want to compute the distribution
        # not through the sample but only through the distributional parameters.
        log_prob = tensor.distribution.log_prob(tensor.detach())
        log_prob = log_prob.sum(dim=tensor.event_dim_indices())

        # Compute the derivative with respect to the distributional parameters through the baseline.
        derivs = []
        for param in get_distr_parameters(
            tensor.distribution, filter_requires_grad=True
        ).values():
            # param.register_hook(hook)
            d_log_prob = storch.grad(
                [log_prob],
                [param],
                create_graph=True,
                grad_outputs=torch.ones_like(log_prob),
            )[0]
            d_output_baseline = storch.grad(
                [output_baseline],
                [param],
                create_graph=True,
                grad_outputs=torch.ones_like(output_baseline),
            )[0]
            derivs.append((param, d_log_prob, d_output_baseline))

        diff = cost_node - output_baseline  # [(...,) + (None,) * d_log_prob.event_dims]
        var_loss = 0.0
        for param, d_log_prob, d_output_baseline in derivs:
            # Compute total derivative with respect to the parameter
            d_param = diff * d_log_prob + d_output_baseline
            # Reduce the plate of this sample in case multiple samples are taken
            d_param = storch.reduce_plates(d_param, plate_names=[tensor.name])
            # Compute backwards from the parameters using its total derivative
            if isinstance(param, storch.Tensor):
                param = param._tensor
            param.backward(d_param._tensor, retain_graph=True)
            # Compute the gradient variance
            variance = (d_param ** 2).sum(d_param.event_dim_indices())
            var_loss += storch.reduce_plates(variance)

        c_phi_params = []

        for param in self.c_phi.parameters(recurse=True):
            if param.requires_grad:
                c_phi_params.append(param)
        d_variance = torch.autograd.grad([var_loss._tensor], c_phi_params)
        for i in range(len(c_phi_params)):
            c_phi_params[i].backward(d_variance[i])
        return None

    def adds_loss(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        return True


class RELAX(MonteCarloMethod):
    """
    Gradient estimator for Bernoulli and Categorical distributions on any function.
    Implements the RELAX estimator from Grathwohl et al, 2018, https://arxiv.org/abs/1711.00123
    and the REBAR estimator from Tucker et al, 2017, https://arxiv.org/abs/1703.07370
    Code inspired by https://github.com/duvenaud/relax/blob/master/pytorch_toy.py
    """

    def __init__(
        self,
        plate_name: str,
        *,
        n_samples: int = 1,
        c_phi: Callable[[torch.Tensor], torch.Tensor] = None,
        in_dim: Dims = None,
        rebar=False,
    ):
        super().__init__(plate_name, n_samples)
        if c_phi:
            self.c_phi = c_phi
        else:
            self.c_phi = Baseline(in_dim)
        self.rebar = rebar
        self.temperature = Parameter(torch.tensor(1.0))
        # TODO: Automatically learn eta
        self.eta = 1.0

    def mc_sample(
        self, distr: Distribution, parents: [storch.Tensor], plates: [Plate]
    ) -> torch.Tensor:
        relaxed_sample = rsample_gumbel(
            distr, self.n_samples, self.temperature, straight_through=False
        )
        # In REBAR, the objective function is evaluated for the Gumbel sample, the conditional Gumbel sample \tilde{z} and the argmax of the Gumbel sample.
        if self.rebar:
            hard_sample = self._discretize(relaxed_sample, distr)
            cond_sample = self._conditional_gumbel_rsample(hard_sample, distr)
            return torch.cat([hard_sample, relaxed_sample, cond_sample], 0)
        else:
            return relaxed_sample
        # sample relaxed if not rebar else sample z then return -> (H(z), z, z|H(z) and n*3

    def post_sample(self, tensor: storch.StochasticTensor) -> Optional[storch.Tensor]:
        if self.rebar:
            # Make sure the r-sampled gumbels don't backpropagate in the normal backward pass
            # TODO: This doesn't work. We need to record the gradients with respect to the cost, as they are used in
            # the estimator. By detaching it, they are no longer recorded. However, not detaching means the normal loss
            # will capture it... Need to rethink this.
            return tensor
        else:
            # Return H(z) for the function evaluation if using RELAX
            return self._discretize(tensor, tensor.distribution)

    def plate_weighting(
        self, tensor: storch.StochasticTensor, plate: storch.Plate
    ) -> Optional[storch.Tensor]:
        # if REBAR: only weight over the true samples, put the relaxed samples to weight 0. This also makes sure
        # that they will not be backpropagated through in the cost backwards pass
        if self.rebar:
            n = int(tensor.n / 3)
            weighting = tensor.new_zeros((3 * n,))
            weighting[:n] = tensor.new_tensor(1.0 / n)
            return weighting
        return super().plate_weighting(tensor, plate)

    def _discretize(self, tensor: torch.Tensor, distr: Distribution) -> torch.Tensor:
        # Adapted from pyro.relaxed_straight_through
        if isinstance(distr, Bernoulli):
            return tensor.round()
        argmax = tensor.max(-1)[1]
        hard_sample = torch.zeros_like(tensor)
        if argmax.dim() < hard_sample.dim():
            argmax = argmax.unsqueeze(-1)
        return hard_sample.scatter_(-1, argmax, 1)

    @deterministic
    def _conditional_gumbel_rsample(
        self, hard_sample: torch.Tensor, distr: Distribution
    ) -> torch.Tensor:
        """
        Conditionally re-samples from the distribution given the hard sample.
        This samples z \sim p(z|b), where b is the hard sample and p(z) is a gumbel distribution.
        """
        # Adapted from torch.distributions.relaxed_bernoulli and torch.distributions.relaxed_categorical
        shape = hard_sample.shape
        probs = (
            distr.probs
            if not isinstance(hard_sample, storch.Tensor)
            else distr.probs._tensor
        )
        probs = clamp_probs(probs.expand_as(hard_sample))
        v = clamp_probs(torch.rand(shape, dtype=probs.dtype, device=probs.device))
        if isinstance(distr, Bernoulli):
            pos_probs = probs[hard_sample == 1]
            v_prime = torch.zeros_like(hard_sample)
            # See https://arxiv.org/abs/1711.00123
            v_prime[hard_sample == 1] = v[hard_sample == 1] * pos_probs + (
                1 - pos_probs
            )
            v_prime[hard_sample == 0] = v[hard_sample == 0] * (
                1 - probs[hard_sample == 0]
            )
            log_sample = (
                probs.log() + probs.log1p() + v_prime.log() + v_prime.log1p()
            ) / self.temperature
            return log_sample.sigmoid()
        # b=argmax(hard_sample)
        b = hard_sample.max(-1).indices
        # b = F.one_hot(b, hard_sample.shape[-1])

        # See https://arxiv.org/abs/1711.00123
        log_v = v.log()
        # i != b (indexing could maybe be improved here, but i doubt it'd be more efficient)
        log_v_b = torch.gather(log_v, -1, b.unsqueeze(-1))
        cond_gumbels = -(-(log_v / probs) - log_v_b).log()
        # i = b
        index_sample = hard_sample.bool()
        cond_gumbels[index_sample] = -(-log_v[index_sample]).log()
        scores = cond_gumbels / self.temperature
        return (scores - scores.logsumexp(dim=-1, keepdim=True)).exp()

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Optional[storch.Tensor]:
        plate = tensor.get_plate(tensor.name)
        if self.rebar:
            hard_sample, relaxed_sample, cond_sample = split(
                tensor, plate, amt_slices=3
            )
            hard_cost, relaxed_cost, cond_cost = split(cost_node, plate, amt_slices=3)

        else:
            hard_sample = self._discretize(tensor, tensor.distribution)
            relaxed_sample = tensor
            cond_sample = self._conditional_gumbel_rsample(
                hard_sample, tensor.distribution
            )

            hard_cost = cost_node
            relaxed_cost = 0.0
            cond_cost = 0.0

        # Input rsampled values into c_phi
        c_phi_relaxed = self.c_phi(relaxed_sample) + relaxed_cost
        c_phi_cond = self.c_phi(cond_sample) + cond_cost

        # Compute log probability of hard sample
        log_prob = tensor.distribution.log_prob(hard_sample)
        log_prob = log_prob.sum(
            dim=list(range(hard_sample.plate_dims, len(log_prob.shape)))
        )

        # Compute the derivative with respect to the distributional parameters through the baseline.
        param = tensor.distribution._param
        # TODO: It should either propagate over the logits or over the probs. Can we know which one is the parameter and
        # which one is computed dynamically?
        # param.register_hook(hook)
        # TODO: Can these be collected in a single torch.autograd.grad call?
        d_log_prob = storch.grad(
            [log_prob],
            [param],
            create_graph=True,
            grad_outputs=torch.ones_like(log_prob),
        )[0]
        d_c_phi_relaxed = storch.grad(
            [c_phi_relaxed],
            [param],
            create_graph=True,
            grad_outputs=torch.ones_like(c_phi_relaxed),
        )[0]
        d_c_phi_cond = storch.grad(
            [c_phi_cond],
            [param],
            create_graph=True,
            grad_outputs=torch.ones_like(c_phi_cond),
        )[0]

        diff = hard_cost - self.eta * c_phi_cond
        # Compute total derivative with respect to the parameter
        d_param = diff * d_log_prob + self.eta * (d_c_phi_relaxed - d_c_phi_cond)
        # Reduce the plate of this sample in case multiple samples are taken
        d_param = storch.reduce_plates(d_param, plate_names=[tensor.name])
        # Compute backwards from the parameters using its total derivative
        if isinstance(param, storch.Tensor):
            param._tensor.backward(d_param._tensor, retain_graph=True)
        else:
            param.backward(d_param._tensor, retain_graph=True)
        # Compute the gradient variance
        variance = (d_param ** 2).sum(d_param.event_dim_indices())
        var_loss = storch.reduce_plates(variance)

        # Minimize variance over the parameters of c_phi and the temperature (should it also minimize eta?)
        c_phi_params = [self.temperature]
        if isinstance(self.c_phi, torch.nn.Module):
            for c_phi_param in self.c_phi.parameters(recurse=True):
                if c_phi_param.requires_grad:
                    c_phi_params.append(c_phi_param)

        d_variance = torch.autograd.grad(
            [var_loss._tensor], c_phi_params, create_graph=self.rebar
        )

        for i in range(len(c_phi_params)):
            c_phi_params[i].backward(d_variance[i])
        return None

    def adds_loss(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        return True


class REBAR(RELAX):
    """
    Gradient estimator for Bernoulli and Categorical distributions on differentiable functions.
    Implements REBAR as a special case of RELAX.
    See Tucker et al, 2017, https://arxiv.org/abs/1703.07370
    """

    def __init__(
        self,
        plate_name: str,
        *,
        c_phi: Callable[[torch.Tensor], torch.Tensor] = None,
        n_samples: int = 1,
    ):
        # Default REBAR does not use external control variate neural network c_phi
        if not c_phi:
            c_phi = lambda x: 0.0
        super().__init__(plate_name, n_samples=n_samples, c_phi=c_phi, rebar=True)



_debug = False

from .wrappers import (
    deterministic,
    stochastic,
    reduce,
    _exception_wrapper,
    _unpack_wrapper,
    ignore_wrapping,
)
from .tensor import Tensor, CostTensor, StochasticTensor, Plate, is_tensor
from .method import *
from .inference import backward, add_cost, reset, denote_independent, gather_samples
from .util import print_graph
from .storch import *
import storch.typing
import storch.nn

import torch

torch.is_tensor = storch.is_tensor
torch.Tensor.to = deterministic(torch.Tensor.to)



import torch


class TensorLike:
    def __torch_function__(func, types, args=(), kwargs=None):
        print("In __torch_function_")


tensor = torch.tensor([1.2, 3.4, 5.6])
tensor.expand_as(TensorLike())
from torch.distributions import Normal

Normal(loc=TensorLike(), scale=1)



_excluded_init = {
    "_debug_has_internal_overlap",
    "_nnpack_available",
    "_use_cudnn_ctc_loss",
    "can_cast",
    "empty",
    "empty_strided",
    "from_file",
    "from_numpy",
    "full",
    "get_default_dtype",
    "get_num_interop_threads",
    "get_num_threads",
    "hamming_window",
    "hann_window",
    "logspace",
    "ones",
    "promote_types",
    "range",
    "tril_indices",
    "triu_indices",
    "set_flush_denormal",
    "set_num_interop_threads",
    "set_num_threads",
    "zeros",
    "compiled_with_cxx11_abi",
}
_unwrap_only_init = {
    "_has_compatible_shallow_copy_type",
    "empty_like",
    "full_like",
    "ones_like",
    "is_distributed",
    "is_floating_point",
    "is_signed",
    "is_same_size",
    "cudnn_is_acceptable",
    "numel",
    "result_type",
    "zeros_like",
}
_exception_init = {
    "allclose",
    "equal",
    "is_complex",
    "is_nonzero",
    "q_per_channel_axis",
    "q_scale",
    "q_zero_point",
    "bernouilli",
    "multinomial",
    "normal",
    "poisson",
}
_to_test_init = {
    "allclose",
    "cudnn_is_acceptable",
    "fbgemm_linear_quantize_weight",
    "numel",
}

_excluded_function = {"get_softmax_dim", "assert_int-or-pair"}

_excluded_tensor = {
    "__getitem__",
    "__setitem__",
    "__class__",
    "__delattr__",
    "__delitem__",
    "__dir__",
    "__format__",
    "__getattribute__",
    "__init_subclass__",
    "__setattr__",
    "__subclasshook__",
}

_exception_tensor = {
    "allclose",
    "bernoulli",
    "bernoulli_",
    "equal",
    "is_complex",
    "is_nonzero",
    "item",
    "normal_" "q_per_channel_axis",
    "q_scale",
    "q_zero_point",
    "random_",
    "stride",
    "to_list",
    "tolist",
    "uniform_",
}

_unwrap_only_tensor = {
    "_make_subclass" "_dimI",
    "_dimV",
    "_is_view",
    "_nnz",
    "dense_dim",
    "get_device",
    "has_names",
    "is_coalesced",
    "is_continguous",
    "is_distributed",
    "is_floating_point",
    "is_pinned",
    "is_same_size",
    "is_set_to",
    "is_signed",
    "nelement",
    "numel",
    "qscheme",
    "size",
    "sparse_dim",
    "storage",
    "storage_offset",
    "new_tensor",
    "new_zeros",
    "new_ones",
}

_to_test_tensor = {"_make_subclass"}

exception_methods = _exception_init.union(_exception_tensor)
excluded_methods = _excluded_tensor.union(_excluded_function).union(_excluded_init)
unwrap_only_methods = _unwrap_only_tensor.union(_unwrap_only_init)

# print(_excluded_init)



from typing import Optional, List

from storch.tensor import Tensor, StochasticTensor, CostTensor, IndependentTensor
import torch
from storch.util import print_graph
import storch


_cost_tensors: [CostTensor] = []
_sampling_methods: [storch.Method] = []


def denote_independent(
    tensor: torch.Tensor,
    dim: int,
    plate_name: str,
    weight: Optional[storch.Tensor] = None,
) -> IndependentTensor:
    """
    Denote the given dimensions on the tensor as being independent, that is, batched dimensions.
    It will automatically put these dimensions to the left.
    :param tensor:
    :param dims:
    :param plate_name: Name of the plate. Reused if called again
    :return:
    """
    if (
        storch.wrappers._context_stochastic
        or storch.wrappers._context_deterministic > 0
    ):
        raise RuntimeError(
            "Cannot create independent tensors within a deterministic or stochastic context."
        )
    if isinstance(tensor, storch.Tensor):
        t_tensor = tensor._tensor
        if dim != 0:
            t_tensor = t_tensor.transpose(dim, 0)
        tensor_name = (
            tensor.name + "_indep_" + plate_name if tensor.name else plate_name
        )
        return IndependentTensor(
            t_tensor, [tensor], tensor.plates, tensor_name, plate_name, weight
        )
    else:
        if dim != 0:
            tensor = tensor.transpose(dim, 0)
        return IndependentTensor(tensor, [], [], plate_name, plate_name, weight)


def gather_samples(
    samples: List[storch.Tensor],
    plate_name: str,
    weight: Optional[storch.Tensor] = None,
) -> IndependentTensor:
    if (
        storch.wrappers._context_stochastic
        or storch.wrappers._context_deterministic > 0
    ):
        raise RuntimeError(
            "Cannot create independent tensors within a deterministic or stochastic context."
        )
    collect_tensors = []
    for sample in samples:
        sample = sample._tensor
        if not sample.shape[0] == 1:
            sample = sample.unsqueeze(0)
        collect_tensors.append(sample)
    cat_tensors = torch.cat(collect_tensors, 0)
    tensor_name = (
        samples[0].name + "_indep_" + plate_name if samples[0].name else plate_name
    )
    return IndependentTensor(
        cat_tensors, samples, samples[0].plates.copy(), tensor_name, plate_name, weight
    )


def add_cost(cost: Tensor, name: str):
    if cost.event_shape != ():
        if cost.event_shape == (1,):
            cost = cost.squeeze(-1)
        else:
            raise ValueError("Can only register cost functions with empty event shapes")
    if not name:
        raise ValueError(
            "No name provided to register cost node. Make sure to register an unique name with the cost."
        )
    cost = CostTensor(cost._tensor, [cost], cost.plates, name)
    if torch.is_grad_enabled():
        storch.inference._cost_tensors.append(cost)
    return cost


def backward(retain_graph=False, debug=False, print_costs=False) -> torch.Tensor:
    """

    :param retain_graph: If set to False, it will deregister the added cost nodes. Should usually be set to False.
    :param debug: Prints debug information on the backwards call.
    :param accum_grads: Saves gradient information in stochastic nodes. Note that this is an expensive option as it
    requires doing O(n) backward calls for each stochastic node sampled multiple times. Especially if this is a
    hierarchy of multiple samples.
    :return:
    """
    costs: [storch.Tensor] = storch.inference._cost_tensors
    if debug:
        print_graph(costs)

    # Sum of averages of cost node tensors
    total_cost = 0.0
    # Sum of losses that can be backpropagated through in keepgrads without difficult iterations
    accum_loss = 0.0

    stochastic_nodes = set()
    # Loop over different cost nodes
    for c in costs:
        # Do not detach the weights when reducing. This is used in for example expectations to weight the
        # different costs.
        reduced_cost = storch.reduce_plates(c, detach_weights=False)

        if print_costs:
            print(c.name, ":", reduced_cost._tensor.item())
        total_cost += reduced_cost
        # Compute gradients for the cost nodes themselves, if they require one.
        if reduced_cost.requires_grad:
            accum_loss += reduced_cost
        for parent in c.walk_parents(depth_first=False):
            # Instance check here instead of parent.stochastic, as backward methods are only used on these.
            if isinstance(parent, StochasticTensor):
                stochastic_nodes.add(parent)
            else:
                continue
            if (
                not parent.requires_grad
                or not parent.sampling_method
                or not parent.sampling_method.adds_loss(parent, c)
            ):
                continue

            # Transpose the parent stochastic tensor, so that its shape is the same as the cost but the event shape, and
            # possibly extra dimensions...?
            parent_tensor = parent._tensor
            reduced_cost = c
            parent_plates = parent.multi_dim_plates()

            # Reduce all plates that are in the cost node but not in the parent node
            for plate in storch.order_plates(c.multi_dim_plates(), reverse=True):
                if plate not in parent_plates:
                    reduced_cost = plate.reduce(reduced_cost, detach_weights=True)

            # Align the parent tensor so that the plate dimensions are in the same order as the cost tensor
            for index_c, plate in enumerate(reduced_cost.multi_dim_plates()):
                index_p = parent_plates.index(plate)
                if index_c != index_p:
                    parent_tensor = parent_tensor.transpose(index_p, index_c)
                    parent_plates[index_p], parent_plates[index_c] = (
                        parent_plates[index_c],
                        parent_plates[index_p],
                    )
            # Add empty (k=1) plates to new parent
            for plate in parent.plates:
                if plate not in parent_plates:
                    parent_plates.append(plate)

            # Create new storch Tensors with different order of plates for the cost and parent
            new_parent = storch.tensor.StochasticTensor(
                parent_tensor,
                [],
                parent_plates,
                parent.name,
                parent.n,
                parent.sampling_method,
                parent.distribution,
                parent._requires_grad,
            )
            # Fake the new parent to be the old parent within the graph by mimicing its place in the graph
            new_parent._parents = parent._parents
            for p, has_link in new_parent._parents:
                p._children.append((new_parent, has_link))
            new_parent._children = parent._children
            cost_per_sample = parent.sampling_method._estimator(
                new_parent, reduced_cost
            )

            if cost_per_sample is not None:
                # The backwards call for reparameterization happens in the
                # backwards call for the costs themselves.
                # Now mean_cost has the same shape as parent.batch_shape
                final_reduced_cost = storch.reduce_plates(
                    cost_per_sample, detach_weights=True
                )
                if final_reduced_cost.ndim == 1:
                    final_reduced_cost = final_reduced_cost.squeeze(0)
                accum_loss += final_reduced_cost

    if isinstance(accum_loss, storch.Tensor) and accum_loss._tensor.requires_grad:
        accum_loss._tensor.backward(retain_graph=retain_graph)

    for s_node in stochastic_nodes:
        if s_node.sampling_method:
            s_node.sampling_method._update_parameters()

    if not retain_graph:
        reset()

    # TODO: How much does accum_loss really say? Should we really keep it? We want to minimize total_cost, anyways.
    return total_cost._tensor  # , accum_loss._tensor


def reset():
    storch.inference._cost_tensors = []
    for method in storch.inference._sampling_methods:
        method.reset()
    storch.inference._sampling_methods = []



from typing import Union, List

import storch
import torch
from storch import deterministic
from torch._C import _infer_size
import torch.nn.functional as F

@deterministic(unwrap=False)
def b_binary_cross_entropy(input: storch.Tensor, target: storch.Tensor, dims: Union[str, List[str]] = None, weight=None, reduction: str = 'mean'):
    r"""Function that measures the Binary Cross Entropy in a batched way
    between the target and the output.

    See :class:`~torch.nn.BCELoss` for details.

    Args:
        input: Tensor of arbitrary shape
        target: Tensor of the same shape as input
        weight (Tensor, optional): a manual rescaling weight
                if provided it's repeated to match input tensor shape
        reduction (string, optional): Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed. Note: :attr:`size_average`
            and :attr:`reduce` are in the process of being deprecated, and in the meantime,
            specifying either of those two args will override :attr:`reduction`. Default: ``'mean'``

    Examples::

        >>> input = torch.randn((3, 2), requires_grad=True)
        >>> target = torch.rand((3, 2), requires_grad=False)
        >>> loss = b_binary_cross_entropy(F.sigmoid(input), target)
        >>> loss.backward()
    """

    if not dims:
        dims = []
    if isinstance(dims, str):
        dims = [dims]

    target = target.expand_as(input)
    unreduced = deterministic(F.binary_cross_entropy)(input, target, weight, reduction="none")

    # unreduced = _loss(input, target, weight)
    indices = list(unreduced.event_dim_indices()) + dims

    if reduction == "mean":
        return storch.mean(unreduced, indices)
    elif reduction == "sum":
        return storch.sum(unreduced, indices)
    elif reduction == "none":
        return unreduced



from storch.method.method import (
    GumbelSoftmax,
    ScoreFunction,
    Infer,
    Method,
    Expect,
    Reparameterization,
    MonteCarloMethod,
)
from storch.method.relax import RELAX, REBAR, LAX
from storch.method.baseline import Baseline, MovingAverageBaseline
from storch.method.sampling import SampleWithoutReplacementMethod
from storch.method.multi_sample_reinforce import ScoreFunctionWOR
from storch.method.unordered import UnorderedSetEstimator



import torch

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
b1 = torch.tensor([True, True, False, True])
c1 = torch.tensor([0.6, 0.6, 0.6])

b2 = torch.tensor([False, False, True, False])
c2 = torch.tensor([0.6])

original_set = torch.Tensor.__setitem__.__get__
original_setitem = torch.Tensor.__setitem__
rewrap = False

import storch


# class wrap:
#     def __init__(self):
#         # self.level = 0
#         self.is_arg_none = True
#         self.arg0 = None
#
#     def __get__(self, *args):
#         # print("getting set")
#         # print(args)
#         self.is_arg_none = args[0] is None
#         self.arg0 = args[0]
#
#         return lambda *args: self.__call__(*args)
#
#     def __call__(self, *args):
#         storch.tensor.level += 1
#         # print("calling set", storch.tensor.level)
#         # print(args)
#         try:
#             if storch.tensor.level == 1 and not self.is_arg_none:
#                 # print("In here")
#                 original_setitem(self.arg0, *args)
#             else:
#                 original_setitem(*args)
#         except TypeError as e:
#             print(storch.tensor.level, e)
#             if storch.tensor.level == 1:
#                 original_setitem(self.arg0, *args)
#         finally:
#             storch.tensor.level -= 1


assert a[b2] == torch.tensor(0.3)
assert (a[b1] == torch.tensor([0.1, 0.2, 0.4])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
a[b2] = c2
assert (a == torch.tensor([0.1, 0.2, 0.6, 0.4])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
torch.Tensor.__setitem__(a, b2, c2)
# a[b2] = c2
assert (a == torch.tensor([0.1, 0.2, 0.6, 0.4])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
a[b1] = c1
assert (a == torch.tensor([0.6, 0.6, 0.3, 0.6])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
a[1] = 0.6
assert (a == torch.tensor([0.1, 0.6, 0.3, 0.4])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
a[1:] = 0.6
assert (a == torch.tensor([0.1, 0.6, 0.6, 0.6])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
a[a > 0.2] = 0.6
assert (a == torch.tensor([0.1, 0.2, 0.6, 0.6])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
a[torch.tensor([2, 3])] = 0.6
assert (a == torch.tensor([0.1, 0.2, 0.6, 0.6])).all()

a = torch.tensor([0.1, 0.2, 0.3, 0.4])
a[1:] = torch.tensor([0.6, 0.7, 0.8])
assert (a == torch.tensor([0.1, 0.6, 0.7, 0.8])).all()

a = torch.tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
assert a[1, 2] == torch.tensor(0.6)
a[1, 2] = 0.7
assert (a == torch.tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.7]])).all()

a = torch.tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
b = torch.tensor([[True, False, True], [False, True, False]])
assert (a[b] == torch.tensor([0.1, 0.3, 0.5])).all()
a[b] = 0.7
assert (a == torch.tensor([[0.7, 0.2, 0.7], [0.4, 0.7, 0.6]])).all()
# print(a[:])
# print("-------------------")
# print(a[b])
# print(torch.Tensor.__getitem__(a, b))



# from typing import Callable, Union
# from torch import Tensor
# from storch import StochasticTensor, stochastic
# # BlackboxTensor
# from storch.typing import AnyBlackboxTensor
#
# class MDP:
#
#     def __init__(self, transition_func: Callable[[Tensor, Tensor], (Tensor, Tensor)], action_func: Callable[[Tensor], StochasticTensor], discount):
#         self.transition_func = stochastic(transition_func)
#         self.action_func = stochastic(action_func)
#         self.discount = discount
#
#     def simulate(self, state: AnyBlackboxTensor, steps=1):
#         for i in range(steps):
#             action_tensor = self.action_func(state)
#             if isinstance(state, BlackboxTensor)
#             if len(action_tensor.batch_links) > 0:
#                 for indices in action_tensor.iterate_batch_indices():
#



import pyro
from pyro.infer import TraceEnum_ELBO, config_enumerate, infer_discrete
import pyro.distributions as dist
import torch

pyro.enable_validation()
pyro.set_rng_seed(0)
def model():
    z = pyro.sample("z", dist.Categorical(torch.ones(5, 5)))
    print('model z = {}'.format(z))

def guide():
    z = pyro.sample("z", dist.Categorical(torch.ones(5, 5)))
    print('guide z = {}'.format(z))

elbo = TraceEnum_ELBO(max_plate_nesting=0)
elbo.loss(model, config_enumerate(guide, "parallel"))


from typing import Optional

import torch
import storch
from storch.method.sampling import (
    SampleWithoutReplacementMethod,
    AncestralPlate,
    log1mexp,
)


class UnorderedSetEstimator(SampleWithoutReplacementMethod):
    def __init__(
        self,
        plate_name: str,
        k: int,
        use_baseline: bool = True,
        exact_integration=False,
        num_int_points=1000,
        a=5.0,
    ):
        super().__init__(plate_name, k)
        self.use_baseline = use_baseline
        self.exact_integration = exact_integration
        self.num_int_points = num_int_points
        self.a = a

    def plate_weighting(
        self, tensor: storch.StochasticTensor, plate: storch.Plate
    ) -> Optional[storch.Tensor]:
        # plates_w_k is a sequence of plates of which one is the input ancestral plate

        # Computes p(s) * R(S^k, s), or the probability of the sample times the leave-one-out ratio.
        # For details, see https://openreview.net/pdf?id=rklEj2EFvB
        # Code based on https://github.com/wouterkool/estimating-gradients-without-replacement/blob/master/bernoulli/gumbel.py
        log_probs = plate.log_probs.detach()

        # Compute integration points for the trapezoid rule: v should range from 0 to 1, where both v=0 and v=1 give a value of 0.
        # As the computation happens in log-space, take the logarithm of the result.
        # N
        v = (
            torch.arange(1, self.num_int_points, out=log_probs._tensor.new())
            / self.num_int_points
        )
        log_v = v.log()

        # Compute log(1-v^{exp(log_probs+a)}) in a numerically stable way in log-space
        # Uses the gumbel_log_survival function from
        # https://github.com/wouterkool/estimating-gradients-without-replacement/blob/master/bernoulli/gumbel.py
        # plates_w_k x N
        g_bound = (
            log_probs[..., None]
            + self.a
            + torch.log(-log_v)[log_probs.plate_dims * (None,) + (slice(None),)]
        )

        # Gumbel log survival: log P(g > g_bound) = log(1 - exp(-exp(-g_bound))) for standard gumbel g
        # If g_bound >= 10, use the series expansion for stability with error O((e^-10)^6) (=8.7E-27)
        # See https://www.wolframalpha.com/input/?i=log%281+-+exp%28-y%29%29
        y = torch.exp(g_bound)
        # plates_w_k x N
        terms = torch.where(
            g_bound >= 10, -g_bound - y / 2 + y ** 2 / 24 - y ** 4 / 2880, log1mexp(y)
        )

        # Compute integrands (without subtracting the special value s)
        # plates x N
        sum_of_terms = storch.sum(terms, plate)
        phi_S = storch.logsumexp(log_probs, plate)
        phi_D_min_S = log1mexp(phi_S)

        # plates x N
        integrand = (
            sum_of_terms
            + torch.expm1(self.a + phi_D_min_S)[..., None]
            * log_v[phi_D_min_S.plate_dims * (None,) + (slice(None),)]
        )

        # Subtract one term the for element that is left out in R
        # Automatically unsqueezes correctly using plate dimensions
        # plates_w_k x N
        integrand_without_s = integrand - terms

        # plates
        log_p_S = integrand.logsumexp(dim=-1)
        # plates_w_k
        log_p_S_without_s = integrand_without_s.logsumexp(dim=-1)

        # plates_w_k
        log_leave_one_out = log_p_S_without_s - log_p_S

        if self.use_baseline:
            # Compute the integrands for the 2nd order leave one out ratio.
            # Make sure to properly choose the indices: We shouldn't subtract the same term twice on the diagonals.
            # k x k
            skip_diag = storch.Tensor(
                1 - torch.eye(plate.n, out=log_probs._tensor.new()), [], [plate]
            )
            # plates_w_k x k x N
            integrand_without_ss = (
                integrand_without_s[..., None, :]
                - terms[..., None, :] * skip_diag[..., None]
            )
            # plates_w_k x k
            log_p_S_without_ss = integrand_without_ss.logsumexp(dim=-1)

            plate.log_snd_leave_one_out = log_p_S_without_ss - log_p_S_without_s

        # Return the unordered set estimator weighting
        return (log_leave_one_out + log_probs).exp().detach()

    def adds_loss(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> bool:
        # We only want to add a loss on the stochastic tensor with the same plate as the cost node.
        # This is because the estimator computes the gradient with the respect to the JOINT log probability.
        # If we would have added the gradient for all stochastic tensors, these would just be duplicates of the same
        # loss being added (ie that gradient would be oversampled)
        for distr_plate in tensor.plates:
            if distr_plate.name == self.plate_name:
                for cost_plate in cost_node.plates:
                    if cost_plate.name == self.plate_name:
                        if cost_plate is distr_plate:
                            return True
                        return False
                raise ValueError(
                    "The given tensor contains an ancestral plate that the cost node doesn't have."
                )
        return False

    def estimator(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> Optional[storch.Tensor]:
        # Note: We automatically multiply with leave-one-out ratio in the plate reduction
        plate = None
        for _p in cost_node.plates:
            if _p.name == self.plate_name:
                plate = _p
                break
        if not self.use_baseline:
            return plate.log_probs * cost_node.detach()

        # Subtract the 'average' cost of other samples, keeping in mind that samples are not independent.
        # plates x k
        baseline = storch.sum(
            (plate.log_probs + plate.log_snd_leave_one_out).exp() * cost_node, plate
        )
        # Make sure the k dimension is recognized as batch dimension
        baseline = storch.Tensor(
            baseline._tensor, [baseline], baseline.plates + [plate]
        )
        advantage = cost_node - baseline
        return plate.log_probs * advantage.detach()



import storch
from torch.distributions import Normal, OneHotCategorical
import torch

torch.is_tensor = storch.is_tensor
torch.manual_seed(0)

# LEGEND OF SHAPES
# These are all different and > 1 to make them identifiable.
# ================
# k: Size of Score Function samples (of which 1 is the kappa dimension)
# |D_yv|: Size of sample from categorical (=4)
# event: event created by repeating in d1
# n1: Plate size of first normal
# n2: Plate size of second normal

k = 4
d_yv = 4
event = 2
plt_n1 = 3
plt_n2 = 2

# Define swr method
swr_method = storch.ScoreFunctionWOR("z", k - 1, biased=True, use_baseline=False)
normal_method1 = storch.ScoreFunction("n1", plt_n1)

l_entropy = torch.tensor([-3.0, -3.0, 2, -2.0], requires_grad=True)
h_entropy = torch.tensor([-0.1, 0.1, 0.05, -0.05], requires_grad=True)

n_params = torch.tensor(0.0, requires_grad=True)


d1 = OneHotCategorical(logits=l_entropy.repeat((event, 1)))
d2 = OneHotCategorical(logits=h_entropy)

dn1 = Normal(n_params, 1.0)

# k x event x |D_yv|
z_1 = swr_method.sample(d1)
# k x |D_yv|
z_2 = swr_method.sample(d2)

print("z1", z_1)
print("z2", z_2)

assert z_1.shape == (min(k, d_yv ** event), event, d_yv)
assert z_2.shape == (min(k, d_yv ** (event + 1)), d_yv)

# n1
n1 = normal_method1(dn1)
print("n1", n1)
_z_1 = z_1 + n1
_z_2 = z_2 + n1

print(_z_1, _z_2)

d3 = OneHotCategorical(logits=_z_1)
d4 = OneHotCategorical(logits=_z_2)

# k x n1 x event x |D_yv|
z_3 = swr_method.sample(d3)
# k x n1 x |D_yv|
z_4 = swr_method.sample(d4)

print("z3", z_3)
print("z4", z_4)

assert z_3.shape == (plt_n1, k, event, d_yv)
assert z_4.shape == (plt_n1, k, d_yv)

normal_method2 = storch.ScoreFunction("n2", plt_n2)
# n2
n2 = normal_method2.sample(dn1)
# n2 x |D_yv|
n_l_entropy = l_entropy + n2.unsqueeze(-1)
d5 = OneHotCategorical(logits=n_l_entropy)

# This is a strange one: although n_l_entropy doesn't include n1, it should be included here, as the ancestral parents
# are dependent on n1!
# k x (n1 x n2 or n2 x n1) x |D_yv|
z_5 = swr_method.sample(d5)

print("z5", z_5)

assert z_5.shape == (plt_n1, plt_n2, k, d_yv) or z_5.shape == (plt_n2, plt_n1, k, d_yv)

d6 = OneHotCategorical(logits=z_3 + z_4.unsqueeze(-2) - z_5.unsqueeze(-2))

z_6 = swr_method.sample(d6)

print("z6", z_6)

assert z_6.shape == (plt_n1, plt_n2, k, event, d_yv) or z_6.shape == (
    plt_n2,
    plt_n1,
    k,
    event,
    d_yv,
)
# Sum the amount of event 1's being true.
cost = torch.sum(z_6[..., 0], -1)
storch.add_cost(cost, "cost")
storch.backward()

# Print what values of z1 are selected in the final sample step. As it has very low entropy, this should be all [0,0,1,0]
# print("final z1", z_6.plates[2].on_unwrap_tensor(z_1))



from __future__ import annotations
import torch
from storch.tensor import Tensor, StochasticTensor, CostTensor
from storch.method.baseline import Baseline
from typing import Union, List, Tuple, Callable
from torch import Size
from torch.distributions import Bernoulli, Categorical, OneHotCategorical

# from storch.seq import BlackboxTensor

_size = Union[Size, List[int], Tuple[int, ...]]

Dims = Union[int, _size]

# AnyBlackboxTensor = Union[BlackboxTensor, torch.Tensor]

DiscreteDistribution = Union[Bernoulli, Categorical, OneHotCategorical]

BaselineFactory = Callable[[StochasticTensor, CostTensor], Baseline]
